"""
cmdtools extension

Class:
    Command:
        Parameters:
            - name: command name

    CommandObject:
        Parameters:
            - name: command name
            - object_: command module or object

    CommandDir:
        Parameters:
            - rootdir: root directory of command files

    CommandModule:
        Parameters:
            - filename: filename or module name

    CommandRunner:
        Parameters:
            - command: CommandObject class to run

    CommandRunnerContainer:
        Parameters:
            - commands: list of CommandObject class to run

    CommandWrapperObject:
        -

    CommandWrapper:
        Parameters:
            - commands: list of CommandWrapperObject class to run

"""

import os
import inspect
import logging
import importlib
import cmdtools
from typing import Any, Callable, List, Union, Optional


class RunnerError(Exception):
    """
    raised when running an unmatched command name
    or some other exceptions
    """


class CommandObject:
    """command object container"""

    def __init__(self, object_: Union[Callable, object], name: str = None):
        self.object = object_

        # try get the command name from the module or class first
        # if no preset name, use command file name as the command name
        self.name = getattr(self.object, "name", name)  # type: str

    @property
    def aliases(self) -> List[str]:
        """get command aliases"""
        return getattr(self.object, "__aliases__", [])

    @property
    def callback(self) -> Optional[Callable]:
        """get command callback"""
        return getattr(self.object, self.name, None)

    @property
    def error_callback(self) -> Optional[Callable]:
        """get command error callback"""
        return getattr(self.object, "error_" + self.name, None)

    @classmethod
    def _checkfunc(cls, func) -> bool:
        """check is object callable"""
        if func is not None and (
            inspect.isfunction(func)
            or inspect.ismethod(func)
            or inspect.iscoroutinefunction(func)
        ):
            return True

        return False

    def is_coroutine(self) -> bool:
        """check whether command is coroutine or not, based by the callbacks"""
        if self.has_callback():
            return inspect.iscoroutinefunction(self.callback)
        if self.has_callback() and self.has_error_callback():
            return inspect.iscoroutinefunction(
                self.callback
            ) and inspect.iscoroutinefunction(self.error_callback)

        # not a coroutine or object does not have command callback
        return False

    def has_callback(self) -> bool:
        """check whether callback is exist or not"""
        return self._checkfunc(self.callback)

    def has_error_callback(self) -> bool:
        """check whether error callback is exist or not"""
        return self._checkfunc(self.error_callback)


class Command(CommandObject):
    """main command class inheritance"""

    def __init__(self, name: str):
        super().__init__(name=name, object_=self)


class CommandWrapperObject:
    """instance of wrapped command object"""

    def __init__(self):
        self.name = None
        self.aliases = []
        self.callback = None
        self.error_callback = None

    def error(self, func: Callable) -> Callable:
        """error handler function wrapper"""
        self.error_callback = func

        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper

    def __call__(self):
        return self.callback(*getattr(self, '_args', ()), **getattr(self, '_kwargs', {}))

    @classmethod
    def _checkfunc(cls, func: Callable) -> bool:
        """check is object callable"""
        if func is not None and (
            inspect.isfunction(func)
            or inspect.ismethod(func)
            or inspect.iscoroutinefunction(func)
        ):
            return True

        return False

    def is_coroutine(self) -> bool:
        """check whether command is coroutine or not, based by the callbacks"""
        if self.has_callback():
            return inspect.iscoroutinefunction(self.callback)
        if self.has_callback() and self.has_error_callback():
            return inspect.iscoroutinefunction(
                self.callback
            ) and inspect.iscoroutinefunction(self.error_callback)

        # not a coroutine or object does not have command callback
        return False

    def has_callback(self) -> bool:
        """check whether callback is exist or not"""
        return self._checkfunc(self.callback)

    def has_error_callback(self) -> bool:
        """check whether error callback is exist or not"""
        return self._checkfunc(self.error_callback)


class CommandWrapper:
    """wrapper for commands (?)"""
    def __init__(self, commands: List[CommandObject] = None):
        if commands is None:
            commands = []

        self.commands = commands

    def command(self, **kwargs) -> Optional[CommandWrapperObject]:
        """base decorator for callbacks"""
        def decorator(obj):
            if (inspect.isfunction(obj), inspect.ismethod(obj), inspect.iscoroutinefunction(obj)).count(True) > 0:
                wrapper = CommandWrapperObject()

                for attr in kwargs.items():
                    setattr(wrapper, attr[0], attr[1])

                wrapper.name = kwargs.get('name', obj.__name__)
                wrapper.aliases = kwargs.get('aliases', [])
                wrapper.callback = obj
                self.commands.append(wrapper)

                def func_wrapper(*args, **kwargs):
                    setattr(wrapper, "_args", args)
                    setattr(wrapper, "_kwargs", kwargs)
                    return wrapper
                return func_wrapper()

            if inspect.isclass(obj):
                if CommandObject in inspect.getmro(obj):
                    self.commands.append(obj())
            else:
                raise TypeError("Cannot assign command for object:", type(obj))
        return decorator

    async def run(self, cmd: Union[cmdtools.Cmd, cmdtools.AioCmd], attrs: dict = None) -> Any:
        """run command instance"""
        if attrs is None:
            attrs = {}

        if not isinstance(cmd, (cmdtools.Cmd, cmdtools.AioCmd)):
            raise TypeError(f"object {type(cmd)} is not a Cmd or AioCmd instance")

        for command in self.commands:
            if command.name == cmd.name or cmd.name in command.aliases:
                args = []
                if command.has_callback():
                    args.append(command.callback)
                else:
                    return logging.error("Command name '%s' has no callback", command.name)
                if command.has_error_callback():
                    args.append(command.error_callback)

                if command.is_coroutine():
                    return await cmd.process_cmd(*args, attrs=attrs)

                return cmd.process_cmd(*args, attrs=attrs)

        raise RunnerError(f"Couln't find command '{cmd.name}'")


class CommandRunner:
    """command single runner class"""

    def __init__(self, command: CommandObject):
        self.command = command

    async def run(self, cmd: Union[cmdtools.Cmd, cmdtools.AioCmd], attrs: dict = None) -> Any:
        """run command from parsed command object"""

        if not isinstance(cmd, (cmdtools.Cmd, cmdtools.AioCmd)):
            raise TypeError(f"object {type(cmd)} is not a Cmd or AioCmd instance")

        if attrs is None:
            attrs = {}

        if self.command.name == cmd.name or cmd.name in self.command.aliases:
            args = []
            if self.command.has_callback():
                args.append(self.command.callback)
            else:
                return logging.error("Command name '%s' has no callback", self.command.name)
            if self.command.has_error_callback():
                args.append(self.command.error_callback)

            if self.command.is_coroutine():
                return await cmd.process_cmd(*args, attrs=attrs)

            return cmd.process_cmd(*args, attrs=attrs)

        raise RunnerError("Command name is invalid")


class CommandRunnerContainer:
    """command runner container class inheritance"""

    def __init__(self, commands: List[CommandObject]):
        self.commands = commands

    async def run(self, cmd: Union[cmdtools.Cmd, cmdtools.AioCmd], attrs: dict = None) -> Any:
        """run command from parsed command object"""

        if not isinstance(cmd, (cmdtools.Cmd, cmdtools.AioCmd)):
            raise TypeError(f"object {type(cmd)} is not a Cmd or AioCmd instance")

        if attrs is None:
            attrs = {}

        for command in self.commands:
            if command.name == cmd.name or cmd.name in command.aliases:
                args = []
                if command.has_callback():
                    args.append(command.callback)
                else:
                    return logging.error("Command name '%s' has no callback", command.name)
                if command.has_error_callback():
                    args.append(command.error_callback)

                if command.is_coroutine():
                    return await cmd.process_cmd(*args, attrs=attrs)

                return cmd.process_cmd(*args, attrs=attrs)

        raise RunnerError(f"Couln't find command '{cmd.name}'")


class CommandModule(CommandRunnerContainer):
    """command module container class"""

    def __init__(self, filename: str, *, load_classes: bool = True):
        self.filename = filename
        self.commands = []

        self.load_classes = load_classes

        self.load_module(load_classes=self.load_classes)
        super().__init__(commands=self.commands)

    def load_module(self, load_classes: bool) -> None:
        """load command classes from a module"""
        if self.filename.endswith(".py"):
            modulestr = (
                self.filename.rsplit(".py", 1)[0].replace(os.sep, ".").strip(".")
            )
            module = importlib.import_module(modulestr)

            if load_classes:
                for obj in dir(module):
                    obj_ = getattr(module, obj, None)

                    if inspect.isclass(obj_) and obj_.__module__ == module.__name__:
                        if CommandObject in inspect.getmro(obj_):
                            self.commands.append(obj_())
            else:
                self.commands.append(CommandObject(module))


class CommandDir(CommandRunnerContainer):
    """command directory container class"""

    def __init__(self, rootdir: str, *, search_tree: bool = False, load_classes: bool = False):
        self.rootdir = rootdir.replace("/", os.sep).replace("\\", os.sep)
        self.commands = []
        self.search_tree: bool = search_tree
        self.load_classes: bool = load_classes

        self.load_commands(search_tree=self.search_tree, load_classes=self.load_classes)
        super().__init__(commands=self.commands)

    def load_commands(self, search_tree: bool, load_classes: bool) -> None:
        """load commands from files inside rootdir"""
        dirs = []
        if search_tree:
            for root in os.walk(self.rootdir):
                dirs.append(root[0])

        if self.rootdir not in dirs:
            dirs.append(self.rootdir)

        for path in dirs:
            for file in os.listdir(path):
                if os.path.isfile(path + os.sep + file) and file.endswith(".py"):
                    module = importlib.import_module(
                        path.replace(os.sep, ".").strip(".")
                        + "."
                        + file.rsplit(".py", 1)[0]
                    )
                    if load_classes is False:
                        self.commands.append(
                            CommandObject(module, file.rsplit(".py", 1)[0])
                        )
                    else:
                        for obj in dir(module):
                            obj_ = getattr(module, obj, None)

                            if inspect.isclass(obj_) and obj_.__module__ == module.__name__:
                                if CommandObject in inspect.getmro(obj_):
                                    self.commands.append(obj_())


def get_command_names(commands: List[CommandObject], get_aliases=False) -> List[str]:
    """get command names from command list of CommandObject"""
    names = []
    for command in commands:
        names.append(command.name)

        if hasattr(command, "aliases") and get_aliases:
            names.extend(command.aliases)

    return names



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'cmdtools'}

packages = \
['cmdtools', 'cmdtools.ext', 'ext']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'cmdtools-py',
    'version': '2.4.9',
    'description': 'command text parser and command processor',
    'long_description': '<div id="headline" align="center">\n  <h1>cmdtools</h1>\n  <p>A module for parsing and processing commands.</p>\n  <a href="https://github.com/HugeBrain16/cmdtools/actions/workflows/python-package.yml">\n    <img src="https://github.com/HugeBrain16/cmdtools/actions/workflows/python-package.yml/badge.svg" alt="tests"></img>\n  </a>\n  <a href="https://pypi.org/project/cmdtools-py">\n    <img src="https://img.shields.io/pypi/dw/cmdtools-py" alt="downloads"></img>\n    <img src="https://badge.fury.io/py/cmdtools-py.svg" alt="PyPI version"></img>\n    <img src="https://img.shields.io/pypi/pyversions/cmdtools-py" alt="Python version"></img>\n  </a>\n  <a href="https://codecov.io/gh/HugeBrain16/cmdtools">\n    <img src="https://codecov.io/gh/HugeBrain16/cmdtools/branch/main/graph/badge.svg?token=mynvRn223H"/>\n  </a>\n  <a href=\'https://cmdtools-py.readthedocs.io/en/latest/?badge=latest\'>\n    <img src=\'https://readthedocs.org/projects/cmdtools-py/badge/?version=latest\' alt=\'Documentation Status\' />\n  </a>\n</div>\n\n## Installation\n\n```\npip install --upgrade cmdtools-py\n```\ninstall latest commit from GitHub  \n```\npip install git+https://github.com/HugeBrain16/cmdtools.git\n```\n## Examples\n\nmore examples [here](https://github.com/HugeBrain16/cmdtools/tree/main/examples)\n\n### Basic example\n\n```py\nimport cmdtools\n\n\ndef ping():\n    print("pong.")\n\n\ncmd = cmdtools.Cmd(\'/ping\')\ncmd.process_cmd(ping)\n```\n  \n### Command with arguments\n\n```py\nimport cmdtools\n\n\ndef give(name, item_name, item_amount):\n    print(f"You gave {item_amount} {item_name}s to {name}")\n\n\n# surround argument that contains whitespaces with quotes\n# set `convert_args` to `True` to automatically convert numbers argument\n\n# this will raise an exception,\n# if the number of arguments provided is less than the number of positional callback parameters.\ncmd = cmdtools.Cmd(\'/give Josh "Golden Apple" 10\', convert_args=True)\n\n# check for command instance arguments data type.\n# format indicates [\'str\',\'str\',\'int\'].\n# integer or float can also match string format, and character \'c\' if the argument only has 1 digit.\n\n# `max_args` set to 3, check the first 3 arguments, the rest will get ignored, \n# otherwise if it set to default,\n# it will raise an exception if the number of arguments is not equal to the number of formats\nif cmd.match_args(\'ssi\', max_args=3):\n    cmd.process_cmd(give)\nelse:\n    print(\'Correct Usage: /give <name: [str]> <item-name: [str]> <item-amount: [int]>\')\n```\n\n## Links\n\nPyPI project: https://pypi.org/project/cmdtools-py  \nSource code: https://github.com/HugeBrain16/cmdtools  \nIssues tracker: https://github.com/HugeBrain16/cmdtools/issues  \nDocumentation: https://cmdtools-py.readthedocs.io/en/latest\n',
    'author': 'HugeBrain16',
    'author_email': 'joshtuck373@gmail.com',
    'maintainer': 'HugeBrain16',
    'maintainer_email': 'joshtuck373@gmail.com',
    'url': 'https://github.com/HugeBrain16/cmdtools',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""A module that does stuff like parsing text commands and processing commands"""

import re
import shlex
import inspect
from typing import Any, Callable, Union, List, Dict

_CVT_FLOAT_PTR = re.compile(r"^[-+]?(\d*[.])\d*$")
_CVT_INT_PTR = re.compile(r"^[-+]?\d+$")


class CmdBaseException(Exception):
    """base exception of this module"""

    def __init__(self, message: str, *args):
        self.message = message
        self.args = args
        super().__init__(self.message)

    def __str__(self):
        return self.message


class ParsingError(Exception):
    """raised when parsing error..."""


class ProcessError(Exception):
    """raised when error occurred during processing commands without error handler"""

    def __init__(self, message: str, exception: Exception):
        self.message = message
        self.exception = exception
        super().__init__(self.message)


class MissingRequiredArgument(CmdBaseException):
    """raised when command's positional argument is missing"""

    def __init__(self, message: str, param: str):
        self.message = message
        self.param = param
        super().__init__(self.message)


class Parser:
    """string prefix parser"""

    def __init__(self, text: str, prefix: str):
        self.text = text
        self.index = -1
        self.prefix = prefix
        self.char = None
        self.args = None

        if self.text:
            self.parse()

    def shift(self) -> None:
        """shift index to next character"""
        self.index += 1
        self.char = self.text[self.index] if self.index < len(self.text) else None

    def parse(self) -> None:
        """parse prefix and get arguments"""
        self.shift()
        space_count = 0
        while self.char is not None:
            if self.char == " ":
                space_count += 1
            else:
                _pref = self.text[:self.index].rstrip()

                if _pref == self.prefix and space_count <= 1:
                    self.args = self.text[self.index:].lstrip()

                    break

            self.shift()


class AttrMan:
    """some kind of attributes manager"""
    def __init__(self, target: Union[Callable, object], **attrs):
        self.target = target
        self.attrs = attrs

        self.dupe_attrs = {}

    def _save_dupes(self) -> None:
        """save existing or duplicate attributes"""
        if not inspect.ismethod(self.target):
            for attr in self.attrs:
                if hasattr(self.target, attr):
                    self.dupe_attrs.update({attr: getattr(self.target, attr)})
        else:
            for attr in self.attrs:
                if hasattr(self.target.__self__, attr):
                    self.dupe_attrs.update({attr: getattr(self.target.__self__, attr)})

    def _clear(self) -> None:
        if not inspect.ismethod(self.target):
            for attr in self.attrs:
                if hasattr(self.target, attr):
                    delattr(self.target, attr)
        else:
            for attr in self.attrs:
                if hasattr(self.target.__self__, attr):
                    delattr(self.target.__self__, attr)

    def set(self) -> None:
        """set attributes to target object"""
        self._save_dupes()

        if not inspect.ismethod(self.target):
            for attr in self.attrs.items():
                setattr(self.target, attr[0], attr[1])
        else:
            for attr in self.attrs.items():
                setattr(self.target.__self__, attr[0], attr[1])

    def reset(self) -> None:
        """reset attributes from target object"""
        self._clear()

        if not inspect.ismethod(self.target):
            for attr in self.dupe_attrs.items():
                setattr(self.target, attr[0], attr[1])
        else:
            for attr in self.dupe_attrs.items():
                setattr(self.target.__self__, attr[0], attr[1])


class Cmd:
    """main class for parsing commands"""

    def __init__(self, command_string: str, prefix="/", max_args: int = 0, convert_args: bool = False):
        self.name = None
        self.args = []
        self.command_string = command_string
        self.convert_args = convert_args
        self.prefix = prefix
        self.max_args = max_args

        # parse command
        res = Parser(self.command_string, self.prefix)
        argres = []
        if res.args is not None:
            argres = shlex.split(res.args)

        if self.max_args == 0:
            self.max_args = len(argres) - 1 if (len(argres) - 1) >= 0 else 0

        if (len(argres) - 1) >= 0 and (len(argres) - 1) > self.max_args:
            raise ParsingError(f"arguments exceeds max arguments: {self.max_args}")

        # insert empty arguments
        while (len(argres) - 1) >= 0 and (len(argres) - 1) < self.max_args:
            argres.append("")

        if argres:
            self.name = argres[0]
            self.args = argres[1:]

            if convert_args:
                self._cvt_cmd()

    def _get_args_type_char(self, max_args: int = 0) -> List[str]:
        """get command arguments data types in char format"""
        argtype = []

        if max_args == 0:
            for arg in self.args[0: len(self.args)]:
                if not arg:
                    continue

                argtype.append(type(arg).__name__[0])  # get type char
        else:
            for arg in self.args[0:max_args]:
                if not arg:
                    continue

                argtype.append(type(arg).__name__[0])  # get type char

        return argtype

    def _cvt_cmd(self) -> None:
        """evaluate literal arguments"""
        cvt = [(_CVT_FLOAT_PTR, float), (_CVT_INT_PTR, int)]

        for idx, arg in enumerate(self.args):
            if not arg:
                break  # empty args

            for cvt_ in cvt:
                res = cvt_[0].match(arg)

                if res:
                    self.args[idx] = cvt_[1](
                        arg
                    )
                    break  # has found the correct data type

    def match_args(self, format_match: str, max_args: int = 0) -> bool:
        """match argument formats, only works with converted arguments"""

        # format example: 'ssf', arguments: ['hell','o',10.0] matched

        if max_args <= 0 and self.max_args > -1:
            max_args = self.max_args

        if not format_match:
            raise ValueError("no format specified")

        format_match = format_match.replace(" ", "")
        format_match = list(format_match)

        argtype = self._get_args_type_char(max_args)

        if len(format_match) != len(argtype):
            raise ValueError("format length is not the same as the arguments length")

        return self._match_args(argtype, format_match)

    def _match_args(self, argtype: List[str], format_match: str) -> bool:
        """match arguments by arguments data types"""
        matched = 0
        for i, arg_type in enumerate(argtype):
            arg_len = len(str(self.args[i]))
            if arg_type in ("i", "f"):
                if format_match[i] == "s":
                    matched += 1  # allow int or float as 's' format
                elif format_match[i] == "c" and arg_len == 1 and arg_type == "i":
                    matched += 1  # and char if only a digit for int
                elif arg_type == format_match[i]:
                    matched += 1
            elif arg_type == "s":
                if format_match[i] == "c" and arg_len == 1:
                    matched += 1
                elif arg_type == format_match[i]:
                    matched += 1

        if matched == len(format_match):
            return True

        return False

    def process_cmd(self, callback: Callable, error_callback: Callable = None, attrs: Dict[str, Any] = None) -> Any:
        """process command..."""
        if attrs is None:
            attrs = {}

        if (
            inspect.isfunction(callback) is False
            and inspect.ismethod(callback) is False
        ):
            raise TypeError("callback is not a function or method")
        if (
            error_callback
            and inspect.isfunction(error_callback) is False
            and inspect.ismethod(callback) is False
        ):
            raise TypeError("error handler callback is not a function")

        if not isinstance(attrs, dict):
            raise TypeError("attributes must be in dict object")

        cman = AttrMan(callback, **attrs)
        ecman = AttrMan(error_callback, **attrs) if error_callback is not None else None

        cman.set()
        if ecman is not None:
            ecman.set()

        ret = None
        try:
            cargspec = inspect.getfullargspec(callback)
            cparams = cargspec.args
            cdefaults = cargspec.defaults

            # remove 'self' or 'cls' from parameters if method
            if inspect.ismethod(callback):
                if len(cparams) > 0:
                    cparams = cparams[1:]

            if cdefaults is None:
                if len(self.args) < len(cparams):
                    raise MissingRequiredArgument(
                        "missing required argument: "
                        + cparams[len(self.args)],
                        param=cparams[len(self.args)],
                    )

            else:

                posargs_length = len(cparams) - len(cdefaults)

                if len(self.args) < posargs_length:
                    raise MissingRequiredArgument(
                        "missing required argument: "
                        + cparams[len(self.args)],
                        param=cparams[len(self.args)],
                    )

            if cargspec.varargs is None:
                ret = callback(*self.args[: len(cparams)])
            else:
                ret = callback(*self.args)

        except Exception as exception:
            if error_callback is None:
                raise ProcessError(
                    "an error occurred during processing callback '"
                    + f"{callback.__name__}()' for command '{self.name}, "
                    + "no error handler callback specified.",
                    exception,
                ) from exception

            error_callback(error=exception)

        cman.reset()
        if ecman is not None:
            ecman.reset()

        return ret

    def __str__(self):
        message = (
            "<"
            + f'Raw: "{self.command_string}", '
            + f'Name: "{self.name}", '
            + f"Args: {self.args}>"
        )

        return message


class AioCmd(Cmd):
    """asynchronous instance of Cmd"""

    async def process_cmd(self, callback: Callable, error_callback: Callable = None, attrs: Dict[str, Any] = None) -> Any:
        """coroutine process cmd"""
        if attrs is None:
            attrs = {}

        if (
            inspect.isfunction(callback) is False
            and inspect.ismethod(callback) is False
        ):
            raise TypeError("callback is not a function or method")
        if (
            error_callback
            and inspect.isfunction(error_callback) is False
            and inspect.ismethod(callback) is False
        ):
            raise TypeError("error handler callback is not a function")

        if not isinstance(attrs, dict):
            raise TypeError("attributes must be in dict object")

        cman = AttrMan(callback, **attrs)
        ecman = AttrMan(error_callback, **attrs) if error_callback is not None else None

        cman.set()
        if ecman is not None:
            ecman.set()

        ret = None
        try:
            cargspec = inspect.getfullargspec(callback)
            cparams = cargspec.args
            cdefaults = cargspec.defaults

            # remove 'self' or 'cls' from parameters if method
            if inspect.ismethod(callback):
                if len(cparams) > 0:
                    cparams = cparams[1:]

            if cdefaults is None:
                if len(self.args) < len(cparams):
                    raise MissingRequiredArgument(
                        "missing required argument: "
                        + cparams[len(self.args)],
                        param=cparams[len(self.args)],
                    )

            else:

                posargs_length = len(cparams) - len(cdefaults)

                if len(self.args) < posargs_length:
                    raise MissingRequiredArgument(
                        "missing required argument: "
                        + cparams[len(self.args)],
                        param=cparams[len(self.args)],
                    )

            if cargspec.varargs is None:
                ret = await callback(*self.args[: len(cparams)])
            else:
                ret = await callback(*self.args)

        except Exception as exception:
            if error_callback is None:
                raise ProcessError(
                    "an error occurred during processing callback '"
                    + f"{callback.__name__}()' for command '{self.name}, "
                    + "no error handler callback specified.",
                    exception,
                ) from exception

            await error_callback(error=exception)

        cman.reset()
        if ecman is not None:
            ecman.reset()

        return ret
